//
// Example of depth peeling in OpenGL via GLSL shaders
//
// Author: Alex V. Boreskoff <alexboreskoff@mtu-net.ru>, <steps3d@narod.ru>
//

#include    "libExt.h"

#ifdef	MACOSX
	#include	<GLUT/glut.h>
#else
	#include	<glut.h>
#endif

#include    <stdio.h>
#include    <stdlib.h>
#include    <assert.h>

#include    "libTexture.h"
#include	"Vector3D.h"
#include	"Vector4D.h"
#include	"GlslProgram.h"
#include	"boxes.h"

Vector3D    eye    ( 7, 5, 7 );       	// camera position
Vector3D	center ( 0, 0, 0 );			// lookAt position
Vector3D	rot    ( 0, 0, 0 );
int			mouseOldX = 0;
int			mouseOldY = 0;
GLuint	    ambMap;                     // ambient texture
GLuint	    diffuseMap;                 // diffuse texture
GLuint		shadowMap;
float       angle = 0;
int			shadowMapSize = 512;
int			layer = 0;
bool		drawDepth = false;
GlslProgram	program;

/////////////////////////////////////////////////////////////////////////////////

void	renderScene ();

void init ()
{
    glClearColor   ( 0.0, 0.0, 0.0, 1.0 );
    glEnable       ( GL_DEPTH_TEST );
    glDepthFunc    ( GL_LEQUAL );

    glHint ( GL_POLYGON_SMOOTH_HINT,         GL_NICEST );
    glHint ( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
}

void    startOrtho ()
{
    glMatrixMode   ( GL_PROJECTION );                   // select the projection matrix
    glPushMatrix   ();                                  // store the projection matrix
    glLoadIdentity ();                                  // reset the projection matrix
                                                        // set up an ortho screen
    glOrtho        ( 0, 512, 0, 512, -1, 1 );
    glMatrixMode   ( GL_MODELVIEW );                    // select the modelview matrix
    glPushMatrix   ();                                  // store the modelview matrix
    glLoadIdentity ();                                  // reset the modelview matrix
}

void    endOrtho ()
{
    glMatrixMode ( GL_PROJECTION );                     // select the projection matrix
    glPopMatrix  ();                                    // restore the old projection matrix
    glMatrixMode ( GL_MODELVIEW );                      // select the modelview matrix
    glPopMatrix  ();                                    // restore the old projection matrix
}

void	renderScene ()
{
    glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    glMatrixMode ( GL_MODELVIEW );
    glLoadIdentity ();
	glPushMatrix ();

    glRotatef    ( rot.x, 1, 0, 0 );
    glRotatef    ( rot.y, 0, 1, 0 );
    glRotatef    ( rot.z, 0, 0, 1 );
	
    glActiveTextureARB ( GL_TEXTURE0_ARB );
    glEnable           ( GL_TEXTURE_2D );
    glBindTexture      ( GL_TEXTURE_2D, diffuseMap );

	glMatrixMode ( GL_MODELVIEW );

	glRotatef       ( 90, 1, 0, 0 );
	glutSolidTeapot ( 1.7 );

	glTranslatef    ( 4, 0, 0 );
	glutSolidTeapot ( 1.7 );

	glTranslatef    ( -4, 0, -3 );
	glutSolidTeapot ( 1.7 );

	glPopMatrix ();
	glFlush     ();
}

void	renderLayer0 ()
{
	renderScene ();
}

void	renderLayer1 ()
{
	renderScene ();	
										// copy depth map into texture
    glActiveTextureARB ( GL_TEXTURE0_ARB );
	glBindTexture      ( GL_TEXTURE_2D, 0 );
	
    glActiveTextureARB ( GL_TEXTURE1_ARB );
	glBindTexture      ( GL_TEXTURE_2D, shadowMap );
	glCopyTexImage2D   ( GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, 0, 0, shadowMapSize, shadowMapSize, 0 );
	glBindTexture      ( GL_TEXTURE_2D, 0 );

	if ( drawDepth )
	{
		glBindTexture      ( GL_TEXTURE_2D,  0 );
	    glActiveTextureARB ( GL_TEXTURE0_ARB );
		glBindTexture      ( GL_TEXTURE_2D, shadowMap );
		glEnable           ( GL_TEXTURE_2D );
		glTexParameteri    ( GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE_ARB, GL_NONE );
		
		glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
		startOrtho ();
		
	    glBegin ( GL_QUADS );

	        glTexCoord2f ( 0, 0 );
	        glVertex2f   ( 0, 0 );

	        glTexCoord2f ( 1,   0 );
	        glVertex2f   ( 511, 0 );

	        glTexCoord2f ( 1, 1 );
	        glVertex2f   ( 511, 511 );

	        glTexCoord2f ( 0, 1 );
	        glVertex2f   ( 0, 511 );
	    glEnd   ();
		
		glBindTexture ( GL_TEXTURE_2D, 0 );
		endOrtho ();
	}
	else
	{
	    glActiveTextureARB ( GL_TEXTURE0_ARB );
	    glBindTexture      ( GL_TEXTURE_2D, diffuseMap );

	    glActiveTextureARB ( GL_TEXTURE1_ARB );
	    glBindTexture      ( GL_TEXTURE_2D, shadowMap );
		glTexParameteri    ( GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE_ARB, GL_COMPARE_R_TO_TEXTURE_ARB );

		program.bind   ();

		renderScene ();						// render scene culling first layer with old depth buffer
		
		program.unbind   ();

	    glActiveTextureARB ( GL_TEXTURE0_ARB );
	    glBindTexture      ( GL_TEXTURE_2D, 0 );
	    glActiveTextureARB ( GL_TEXTURE1_ARB );
	    glBindTexture      ( GL_TEXTURE_2D, 0 );
	}
}

void	renderLayer2 ()
{
	renderLayer1 ();					// now we have in shadowMap current depth buffer

										// copy depth map into texture
    glActiveTextureARB ( GL_TEXTURE1_ARB );
	glBindTexture      ( GL_TEXTURE_2D, shadowMap );
	glCopyTexImage2D   ( GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, 0, 0, shadowMapSize, shadowMapSize, 0 );

	
    glActiveTextureARB ( GL_TEXTURE0_ARB );
    glEnable           ( GL_TEXTURE_2D );
    glBindTexture      ( GL_TEXTURE_2D, diffuseMap );

    glActiveTextureARB ( GL_TEXTURE1_ARB );
    glEnable           ( GL_TEXTURE_2D );
    glBindTexture      ( GL_TEXTURE_2D, shadowMap );

	program.bind   ();

	
	renderScene ();						// render scene culling first layer with old depth buffer

	program.unbind   ();
}

void	renderLayer3 ()
{
	renderLayer2 ();					// now we have in shadowMap current depth buffer

										// copy depth map into texture
    glActiveTextureARB ( GL_TEXTURE1_ARB );
	glBindTexture      ( GL_TEXTURE_2D, shadowMap );
	glCopyTexImage2D   ( GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, 0, 0, shadowMapSize, shadowMapSize, 0 );

	
    glActiveTextureARB ( GL_TEXTURE0_ARB );
    glEnable           ( GL_TEXTURE_2D );
    glBindTexture      ( GL_TEXTURE_2D, diffuseMap );

    glActiveTextureARB ( GL_TEXTURE1_ARB );
    glEnable           ( GL_TEXTURE_2D );
    glBindTexture      ( GL_TEXTURE_2D, shadowMap );

	program.bind   ();

	renderScene ();						// render scene culling first layer with old depth buffer

	program.unbind   ();
}

void display ()
{
	if ( layer == 0 )
		renderLayer0 ();
	else
	if ( layer == 1 )
		renderLayer1 ();
	else
	if ( layer == 2 )
		renderLayer2 ();
	else
		renderLayer3 ();
		
    glutSwapBuffers ();
}

void reshape ( int w, int h )
{
    glViewport     ( 0, 0, (GLsizei)w, (GLsizei)h );
   	glMatrixMode   ( GL_PROJECTION );
	glLoadIdentity ();

    gluPerspective ( 60.0, (GLfloat)w/(GLfloat)h, 3, 20.0 );
    gluLookAt      ( eye.x, eye.y, eye.z,    		// eye
                     center.x, center.y, center.z,  // center
                     0, 0, 1 );              		// up

	glMatrixMode   ( GL_MODELVIEW );
    glLoadIdentity ();
}

void motion ( int x, int y )
{
    rot.y -= ((mouseOldY - y) * 180.0f) / 200.0f;
    rot.z -= ((mouseOldX - x) * 180.0f) / 200.0f;
    rot.x  = 0;

    if ( rot.z > 360 )
		rot.z -= 360;

	if ( rot.z < -360 )
		rot.z += 360;

    if ( rot.y > 360 )
		rot.y -= 360;

	if ( rot.y < -360 )
        rot.y += 360;

    mouseOldX = x;
    mouseOldY = y;

	glutPostRedisplay ();
}

void mouse ( int button, int state, int x, int y )
{
    if ( state == GLUT_DOWN )
    {
        mouseOldX = x;
        mouseOldY = y;
	}
}

void key ( unsigned char key, int x, int y )
{
    if ( key == 27 || key == 'q' || key == 'Q' )        // quit requested
        exit ( 0 );
		
	if ( key == '0' )
		layer = 0;
	else
	if ( key == '1' )
		layer = 1;
	else
	if ( key == '2' )
		layer = 2;
	else
	if ( key == '3' )
		layer = 3;
		
	if ( key == ' ' )
		drawDepth = !drawDepth;
		
    glutPostRedisplay ();		
}

int main ( int argc, char * argv [] )
{
                                // initialize glut
    glutInit            ( &argc, argv );
    glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH );
    glutInitWindowSize  ( shadowMapSize, shadowMapSize );

                                // create window
    glutCreateWindow ( "OpenGL depth peeling demo" );

                                // register handlers
    glutDisplayFunc  ( display );
    glutReshapeFunc  ( reshape );
    glutKeyboardFunc ( key     );
	glutMouseFunc    ( mouse   );
	glutMotionFunc   ( motion  );

    init           ();
    initExtensions ();

	assertExtensionsSupported ( "GL_ARB_shadow GL_ARB_depth_texture" );

	if ( !GlslProgram :: isSupported () )
	{
		printf ( "GLSL not supported.\n" );

		return 1;
	}

	if ( !program.loadShaders ( "depth-peel.vsh", "depth-peel.fsh" ) )
	{
		printf ( "Error loading shaders:\n%s\n", program.getLog ().c_str () );

		return 2;
	}
	
    diffuseMap  = createTexture2D ( true,  "../../Textures/oak.bmp" );

								// create depth texture
	glGenTextures   ( 1, &shadowMap );
	glBindTexture   ( GL_TEXTURE_2D, shadowMap );
	glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
	glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
	glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );

	glTexParameteri  ( GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE_ARB, GL_COMPARE_R_TO_TEXTURE_ARB );
	glTexParameteri  ( GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC_ARB, GL_LEQUAL );

	glTexImage2D     ( GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, shadowMapSize, shadowMapSize, 0,
                       GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, NULL );

    program.bind ();
	program.setTexture ( "decalMap",  0 );
	program.setTexture ( "depthMap",  1 );
	program.unbind ();
					   
    glutMainLoop ();

    return 0;
}
